/*
	garith.c -- A GNOME Arithmetic test
		Version 2.3
*/

/*

This and all other files included with this distribution are
Copyright 1997 Micah Stetson.  They are fully redistributable
under the terms of the GNU General Public License. For details,
please see the COPYING file which you should have recieved with
this file.

*/

/*
	This program has probably not been put together in the
	most efficient way. If you see areas in which it may be
	improved, feel free to suggest such improvement to me or
	if you would rather improve it yourself, just send
	me a patch. I will, more than likely, include these
	suggestions/patches in the main distribution and give you
	credit for them in the README and HISTORY files.
*/

#include <gtk/gtk.h>		/* GTK stuff */
#include <gnome.h>		/* GNOME stuff */
#include <stdlib.h>
#include <time.h>		/* we use time as seed for rand() */

/*
	Global variables that need to be accessed by multiple
	functions.  If I/you can figure out a good way to eliminate
	some/all of these, it ought to be done.
*/

	GtkWidget	*main_window;
	GtkWidget	*pref_window;
	GtkWidget	*pixmap;
	GtkWidget	*right_pct;
	GtkWidget	*complete_num;
	GtkWidget	*speed_num;
	GtkWidget	*probnum1;
	GtkWidget	*probop;
	GtkWidget	*probnum2;
	GtkWidget	*difficulty;

/*
	XPM variables to hold XPM data for X Arith logo,
	the 'RIGHT!' pixmap, and the 'WRONG!' pixmap.
	The actual data is at the end of this file.
	I think I'm going to move the pictures out of the
	source and into their own files, this will be easier
	to maintain.  I need a better G Arith logo, this one
	was made when it was called X Arith and used XForms.
	XForms has a darker default background color on which
	the white text of the logo shows up quite well; it
	doesn't show up well at all though in GTK so I need to
	fix it.
*/

static gchar * garith_logo[];
static gchar * right_xpm[];
static gchar * wrong_xpm[];

/*
	Global variables for the operation, the two 
	factors (or whatever), how many correct 
	answers were given, problems completed, the 
	integer that the factors must be less than, 
	and the starting time of the current test 
	(reset_score() sets this).  I haven't finished
	implementing the timer; I may use a gtimer for
	this.
*/

gulong		num1, num2;
guint		right = 0, done = 0;
gldouble	less_than;
gchar		op;
time_t		starttime;

/* Pointer to the program name, for error messages */
char *progname;

/* pixmaps */
enum {
	GARITH,
	RIGHT,
	WRONG
};

/*
	This function sets the main pixmap to one of the ones in the
	enum declaration above.

	If return is TRUE, it returns the pixmap, otherwise it returns
	the mask.

	Much of this was taken from get_monitor_preview_widget() in
	main.c of GNOME's desktop_properties program, thanks.
*/

GdkPixmap* get_pixmap(gint pmap, GdkPixmap **msk) {
	GdkPixmap *xpm, *mask;
	GtkStyle  *style;
	gchar *f = NULL;

	switch (pmap) {
		case GARITH:
			f = gnome_datadir_file("pixmaps/garith_logo.xpm");
			break;
		case RIGHT:
			f = gnome_datadir_file("pixmaps/garith_right.xpm");
			break;
		case WRONG:
			f = gnome_datadir_file("pixmaps/garith_wrong.xpm");
			break;
	}

	if(!f)
		g_warning("%s: Pixmap not found", progname);

	style = gtk_widget_get_style (main_window);
	xpm = gdk_pixmap_create_from_xpm(main_window->window, &mask,
		&style->bg [GTK_STATE_NORMAL], f);

	*msk = mask;
	return xpm;
}

/*
	This function sets problem characteristics and displays
	a randomly generated problem. It also handles display of
	the options window. This should probably be moved to a
	different function that would handle all windows.  But as
	the options window is the only subwindow at this point and
	as this function handles all interaction with that window,
	it stays here for now.  I may move it in the next version.
*/

void set_problem(GtkWidget *widget, gpointer *data) {
	gint		digits, c;
	gchar		string[10];

	switch((gint) data) {
		case 0:	/* New Problem */
			/* I don't like this below, but I don't know
				a more efficient way to do it.  This
				(or something that serves the same
				purpose) is necessary in order to
				ensure that our answers are always
				integers.  Perhaps the higher levels
				should have floating point answers,
				but at least the lower levels (3 or 4
				and below) should not, IMHO. */
			if (op == '/') {
				num1 = (gulong) (less_than * rand()
					/ (RAND_MAX+1.0));
				if (num1 > (less_than / 10.0))
					num2 = 1 + (gulong) (10.0 * rand()
						/ (RAND_MAX+1.0));
				else if (num1 > (less_than / 100.0))
					num2 = 1 + (gulong) (100.0 * rand()
						/ (RAND_MAX+1.0)); 
				else if (num1 > (less_than / 1000.0))
					num2 = 1 + (gulong) (1000.0 * rand()
						/ (RAND_MAX+1.0)); 
				else if (num1 > (less_than / 10000.0))
					num2 = 1 + (gulong) (10000.0 * rand()
						/ (RAND_MAX+1.0)); 
				else if (num1 > (less_than / 100000.0))
					num2 = 1 + (gulong) (100000.0 * rand()
						/ (RAND_MAX+1.0)); 
				else if (num1 > (less_than / 1000000.0))
					num2 = 1 + (gulong) (1000000.0
						* rand() / (RAND_MAX+1.0)); 
				else if (num1 > (less_than / 10000000.0))
					num2 = 1 + (gulong) (10000000.0
						* rand() / (RAND_MAX+1.0)); 
				else if (num1 >= 1)
					num2 = 1 + (gulong) (100000000.0
						* rand() / (RAND_MAX+1.0));
				else if (num1 == 0)
					num2 = 1 + (gulong) (less_than
						* rand() / (RAND_MAX+1.0));
			} else {
				num1 = (gulong) (less_than * rand()
					/ (RAND_MAX+1.0));
				num2 = (gulong) (less_than * rand()
					/ (RAND_MAX+1.0));
			}

			if (op == '/') {
				sprintf(string, "%lu", (num1 * num2));
				gtk_label_set(GTK_LABEL (probnum1), string);

				sprintf(string, "%lu", num2);
				gtk_label_set(GTK_LABEL (probnum2), string);
			} else if (op == '-' && num1 < num2) {
				sprintf(string, "%lu", num2);
				gtk_label_set(GTK_LABEL (probnum1), string);

				sprintf(string, "%lu", num1);
				gtk_label_set(GTK_LABEL (probnum2), string);
			} else {
				sprintf(string, "%lu", num1);
				gtk_label_set(GTK_LABEL (probnum1), string);

				sprintf(string, "%lu", num2);
				gtk_label_set(GTK_LABEL (probnum2), string);
			}

			sprintf(string, "%c", op);
			gtk_label_set(GTK_LABEL (probop), string);

			break;
		case 1:		/* Digits */
			c = 0;
			digits = (gint) GTK_ADJUSTMENT (widget)->value;

			for(less_than = 1.0; c < digits; c++) {
				less_than *= 10.0;
			}
			/* The first arg here is meaningless */
			set_problem(probnum1, 0);
			break;
		case 2:		/* Addition */
			op = '+';
			/* Again, arg1 below is meaningless */
			set_problem(probop, 0);
			break;
		case 3:		/* Subtraction */
			op = '-';
			/* Again, arg1 below is meaningless */
			set_problem(probop, 0);
			break;
		case 4:		/* Multiplication */
			op = '*';
			/* Again, arg1 below is meaningless */
			set_problem(probop, 0);
			break;
		case 5:		/* Division */
			op = '/';
			/* Again, arg1 below is meaningless */
			set_problem(probop, 0);
			break;
		case 6:		/* Options Button */
			gtk_widget_show(pref_window);
			break;
		case 7:		/* Done Button */
			/* Closes pref_window */
			gtk_widget_hide(pref_window);

			break;
	}	

	return;
}

/*
	This function is called when the user enters a solution
	to the displayed problem.  It checks the answer, increments
	the number stored in right (if the user was right), displays
	a pixmap indicating whether the user's answer was correct or
	incorrect, increments the number of completed problems (done), 
	recalculates the percentage of correct problems, updates the 
	score display, and calls set_problem to display a new problem.
*/

void solve_prob(GtkWidget *widget) {
	gint		answer = 0;
	gint		solution = 0;
	time_t		now;
	gdouble		elapsed;
	gchar		string[10];
	GdkPixmap	*xpm, *mask;

	/* Get user's answer */

	answer = atoi(gtk_entry_get_text(GTK_ENTRY (widget)));

	/* Figure out the real answer */

	switch(op) {
		case '+':
			solution = num1 + num2;
			break;
		case '-':
			solution = abs(num1 - num2);
			break;
		case '*':
			solution = num1 * num2;
			break;
		case '/':
			solution = num1;
			break;
	}

	if (answer == solution) {	/* User was right */
		xpm = get_pixmap(RIGHT, &mask);
		gtk_pixmap_set(GTK_PIXMAP (pixmap), xpm, mask);
		right++;

	} else {			/* User was wrong */
		xpm = get_pixmap(WRONG, &mask);
		gtk_pixmap_set(GTK_PIXMAP (pixmap), xpm, mask);
	}

	done++;
	sprintf(string, "%d", done);
	gtk_label_set(GTK_LABEL (complete_num), string);

	sprintf(string, "%d%%",
		(int) (100.0 * ((float) right / (float) done)));
	gtk_label_set(GTK_LABEL (right_pct), string);

	/* Figure average seconds per problem, and display */
	now = time(NULL);
	/* This does not work, need to read the difftime man page
		again.  Of course, I won't need to if I use a
		gtimer; need to read about that. */
	elapsed = difftime(starttime, now);
	sprintf(string, "%f", (float) elapsed / (float) done);

	set_problem(widget, 0);		/* first arg meaningless */	
	gtk_entry_set_text(GTK_ENTRY (widget), "");

	return;
}

/*
	This function resets the displays right_num and wrong_num to
	0 and calls set_problem to display a new one.  Also resets
	global variables pertaining to score.
*/

void reset_score(GtkWidget *widget, gpointer *data) {
	gchar		string[5];
	GdkPixmap	*xpm, *mask;

	/* Reset number right */
	right = 0;
	sprintf(string, "%d%%", 100);
	gtk_label_set(GTK_LABEL (right_pct), string);

	/* Reset number done */
	done = 0;
	sprintf(string, "%d", 0);
	gtk_label_set(GTK_LABEL (complete_num), string);

	/* Reset timer display */
	/* not gotten to this yet */

	/* Reset pixmap to G Arith Logo */
	xpm = get_pixmap(GARITH, &mask);
	gtk_pixmap_set(GTK_PIXMAP (pixmap), xpm, mask);

	/* Produce new problem */
	set_problem(widget, (gpointer) 0);

	/* Reset timer */
	starttime = time(NULL);

	return;
}

void show_help(GtkWidget *widget, gpointer *data)
{
	return;
}

void show_high(GtkWidget *widget, gpointer *data)
{
	return;
}

void quit_confirm(GtkWidget *widget, gpointer *data)
{
	gtk_main_quit();
}

/*
	This function creates all the widgets used in the program.
	It is separate because it makes the source more readable for me.
*/

gint make_widgets(void)
{
	GtkWidget	*widget1;
	GtkWidget	*widget2;
	GtkWidget	*box1;
	GtkWidget	*box2;
	GtkWidget	*box3;
	GtkWidget	*box4;
	GtkWidget	*box5;
	GtkWidget	*entry;
	GtkObject	*adjustment;
	GdkPixmap	*xpm, *mask;

	GtkMenuFactory	*factory;
	gint		entrynum;
	GtkMenuEntry	menu_entries[] = 
	{
		{"File/New Game", "<control>N", (GtkMenuCallback) reset_score, NULL},
		{"File/Preferences...", "<control>P", (GtkMenuCallback) set_problem, (gpointer) 6},
		{"File/High Scores...", "<control>H", (GtkMenuCallback) show_high, (gpointer) 0},
		{"File/<separator>", NULL, NULL, NULL},
		{"File/Quit...", "<control>Q", (GtkMenuCallback) quit_confirm, (gpointer) 0},
		{"Help/About...", "<control>A", (GtkMenuCallback) show_help, (gpointer) 0}
	};

	entrynum = sizeof(menu_entries) / sizeof(GtkMenuEntry);

	/* Begin main_window */
	main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_title(GTK_WINDOW (main_window),
			"GNOME Arithmetic");
		gtk_signal_connect(GTK_OBJECT (main_window), "destroy",
					(GtkSignalFunc) quit_confirm, NULL);

	/* Main vbox */
	box1 = gtk_vbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER (main_window), box1);
		gtk_widget_show(box1);

	/* Menu */
	factory = gtk_menu_factory_new(GTK_MENU_FACTORY_MENU_BAR);
		gtk_menu_factory_add_entries(factory, menu_entries, entrynum);
		gtk_window_add_accelerator_table(GTK_WINDOW(main_window), factory->table);
		gtk_box_pack_start(GTK_BOX (box1), factory->widget, FALSE, FALSE, 0);
		gtk_widget_show(factory->widget);

	/* Problem/Solution/Score/Logo Area */
	box2 = gtk_hbox_new(TRUE, 10);
		gtk_container_border_width (GTK_CONTAINER (box2), 10);
		gtk_box_pack_start(GTK_BOX (box1), box2, FALSE, FALSE, 0);
		gtk_widget_show(box2);
	/* Problem/Solution Area */
	widget1 = gtk_frame_new("Problem");
		gtk_box_pack_start(GTK_BOX (box2), widget1, TRUE, FALSE, 0);
		gtk_widget_show(widget1);
	box3 = gtk_vbox_new(TRUE, 0);
		gtk_container_border_width (GTK_CONTAINER (box3), 10);
		gtk_container_add(GTK_CONTAINER (widget1), box3);
		gtk_widget_show(box3);
	box4 = gtk_hbox_new(FALSE, 10);
		gtk_box_pack_start(GTK_BOX (box3), box4, TRUE, FALSE, 0);
		gtk_widget_show(box4);
	probnum1 = gtk_label_new("0");
		gtk_box_pack_end(GTK_BOX (box4), probnum1, FALSE, FALSE, 0);
		gtk_widget_show(probnum1);
	box4 = gtk_hbox_new(FALSE, 10);
		gtk_box_pack_start(GTK_BOX (box3), box4, TRUE, FALSE, 0);
		gtk_widget_show(box4);
	probnum2 = gtk_label_new("0");
		gtk_box_pack_end(GTK_BOX (box4), probnum2, FALSE, FALSE, 0);
		gtk_widget_show(probnum2);
	probop = gtk_label_new("+");
		gtk_box_pack_end(GTK_BOX (box4), probop, FALSE, FALSE, 0);
		gtk_widget_show(probop);
	widget1 = gtk_hseparator_new();
		gtk_box_pack_start(GTK_BOX (box3), widget1, TRUE, FALSE, 0);
		gtk_widget_show(widget1);
	entry = gtk_entry_new();
		gtk_signal_connect_object(GTK_OBJECT (entry), "activate",
			(GtkSignalFunc) solve_prob, GTK_OBJECT (entry));
		gtk_box_pack_start(GTK_BOX (box3), entry, TRUE, FALSE, 0);
		gtk_widget_show(entry);

	box3 = gtk_vbox_new(FALSE, 10);
		gtk_box_pack_start(GTK_BOX (box2), box3, TRUE, TRUE, 0);
		gtk_widget_show(box3);

	/* Score display */
	widget1 = gtk_frame_new("Score");
		gtk_box_pack_start(GTK_BOX (box3), widget1, FALSE, FALSE, 0);
		gtk_widget_show(widget1);
	box4 = gtk_vbox_new(TRUE, 10);
		gtk_container_border_width (GTK_CONTAINER (box4), 10);
		gtk_container_add(GTK_CONTAINER (widget1), box4);
		gtk_widget_show(box4);
	box5 = gtk_hbox_new(TRUE, 10);
		gtk_box_pack_start(GTK_BOX (box4), box5, TRUE, FALSE, 0);
		gtk_widget_show(box5);
	widget1 = gtk_frame_new("Right");
		gtk_frame_set_label_align(GTK_FRAME (widget1), 0.5, 0.0);
		gtk_box_pack_start(GTK_BOX (box5), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	right_pct = gtk_label_new("100%");
		gtk_container_add(GTK_CONTAINER (widget1), right_pct);
		gtk_widget_show(right_pct);
	widget1 = gtk_frame_new("Completed");
		gtk_frame_set_label_align(GTK_FRAME (widget1), 0.5, 0.0);
		gtk_box_pack_start(GTK_BOX (box5), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	complete_num = gtk_label_new("0");
		gtk_container_add(GTK_CONTAINER (widget1), complete_num);
		gtk_widget_show(complete_num);
	widget1 = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
		gtk_box_pack_start(GTK_BOX (box4), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	widget2 = gtk_frame_new("Avg. Speed");
		gtk_frame_set_label_align(GTK_FRAME (widget2), 0.5, 0.0);
		gtk_container_add(GTK_CONTAINER (widget1), widget2);
		gtk_widget_show(widget2);
	speed_num = gtk_label_new("0");
		gtk_container_add(GTK_CONTAINER (widget2), speed_num);
		gtk_widget_show(speed_num);

	/* Logo/Right/Wrong pixmap and its frame */
	widget1 = gtk_frame_new(NULL);
		gtk_box_pack_start(GTK_BOX (box3), widget1, FALSE, FALSE, 0);
		gtk_widget_show(widget1);
	xpm = get_pixmap(GARITH, &mask);
	pixmap = gtk_pixmap_new(xpm, mask);
		gtk_container_add(GTK_CONTAINER (widget1), pixmap);
		gtk_widget_show(pixmap);

	/* End main_window */

	/* Begin pref_window */
	pref_window = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_window_set_title(GTK_WINDOW (pref_window),
			"G Arith: Preferences");
		gtk_container_border_width (GTK_CONTAINER (pref_window), 10);
		gtk_signal_connect_object(GTK_OBJECT (pref_window), "destroy",
					(GtkSignalFunc) gtk_widget_hide,
					GTK_OBJECT (pref_window));

	box1 = gtk_vbox_new(FALSE, 10);
		gtk_container_add(GTK_CONTAINER (pref_window), box1);
		gtk_widget_show(box1);

	box2 = gtk_hbox_new(FALSE, 10);
		gtk_box_pack_start(GTK_BOX (box1), box2, FALSE, FALSE, 0);
		gtk_widget_show(box2);

	widget1 = gtk_notebook_new();
		gtk_notebook_set_tab_pos(GTK_NOTEBOOK (widget1), GTK_POS_TOP);
		gtk_box_pack_start(GTK_BOX (box2), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);

	box2 = gtk_hbox_new(TRUE, 10);
		gtk_container_border_width (GTK_CONTAINER (box2), 10);
	widget2 = gtk_label_new("Game Settings");
		gtk_notebook_append_page(GTK_NOTEBOOK (widget1), box2, widget2);
		gtk_widget_show(box2);
		gtk_widget_show(widget2);

	box3 = gtk_vbox_new(TRUE, 10);
		gtk_container_border_width (GTK_CONTAINER (box3), 10);
	widget2 = gtk_label_new("Colors/Fonts");
		gtk_notebook_append_page(GTK_NOTEBOOK (widget1), box3, widget2);
		gtk_widget_show(box3);
		gtk_widget_show(widget2);

	box4 = gtk_vbox_new(FALSE, 10);
		gtk_box_pack_start(GTK_BOX (box2), box4, TRUE, FALSE, 0);
		gtk_widget_show(box4);

	adjustment = gtk_adjustment_new(2.0, 1.0, 9.0, 1.0, 1.0, 1.0);
		gtk_signal_connect(GTK_OBJECT (adjustment), "value_changed",
			(GtkSignalFunc) set_problem, (gpointer) 1);

	difficulty = gtk_vscale_new(GTK_ADJUSTMENT (adjustment));
		gtk_scale_set_digits(GTK_SCALE (difficulty), 0);
		gtk_box_pack_start(GTK_BOX (box4), difficulty, FALSE, FALSE, 0);
		gtk_widget_show(difficulty);
	widget1 = gtk_label_new("Difficulty");
		gtk_box_pack_start(GTK_BOX (box4), widget1, FALSE, FALSE, 0);
		gtk_widget_show(widget1);

	box4 = gtk_vbox_new(TRUE, 10);
		gtk_box_pack_start(GTK_BOX (box2), box4, TRUE, FALSE, 0);
		gtk_widget_show(box4);

	widget1 = gtk_radio_button_new_with_label(NULL, "Addition");
		gtk_signal_connect(GTK_OBJECT (widget1), "clicked",
			(GtkSignalFunc) set_problem, (gpointer) 2);
		gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (widget1), TRUE);
		gtk_box_pack_start(GTK_BOX (box4), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	widget1 = gtk_radio_button_new_with_label(
			gtk_radio_button_group(GTK_RADIO_BUTTON (widget1)),
			"Subtraction");
		gtk_signal_connect(GTK_OBJECT (widget1), "clicked",
			(GtkSignalFunc) set_problem, (gpointer) 3);
		gtk_box_pack_start(GTK_BOX (box4), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	widget1 = gtk_radio_button_new_with_label(
		gtk_radio_button_group(GTK_RADIO_BUTTON (widget1)),
			"Multiplication");
		gtk_signal_connect(GTK_OBJECT (widget1), "clicked",
			(GtkSignalFunc) set_problem, (gpointer) 4);
		gtk_box_pack_start(GTK_BOX (box4), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);
	widget1 = gtk_radio_button_new_with_label(
		gtk_radio_button_group(GTK_RADIO_BUTTON (widget1)),
			"Division");
		gtk_signal_connect(GTK_OBJECT (widget1), "clicked",
			(GtkSignalFunc) set_problem, (gpointer) 5);
		gtk_box_pack_start(GTK_BOX (box4), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);

	widget1 = gtk_label_new("Not yet implemented.");
		gtk_box_pack_start(GTK_BOX (box3), widget1, TRUE, TRUE, 0);
		gtk_widget_show(widget1);

	box2 = gtk_hbox_new(FALSE, 10);
		gtk_box_pack_end(GTK_BOX (box1), box2, FALSE, FALSE, 0);
		gtk_widget_show(box2);

	widget1 = gtk_button_new_with_label("Close");
		gtk_signal_connect(GTK_OBJECT (widget1), "clicked",
			(GtkSignalFunc) set_problem, (gpointer) 7);
		gtk_box_pack_end(GTK_BOX (box2), widget1, FALSE, FALSE, 0);
		GTK_WIDGET_SET_FLAGS (widget1, GTK_CAN_DEFAULT);
		gtk_widget_grab_default(widget1);
		gtk_widget_show(widget1);

	/* End pref_window */

	return 0;
}

/*
	main(): Where all the fun begins.
*/

gint main(int argc, char *argv[])
{
	time_t now;
	gint seed;

	/* gnome_init() is called by all gnome apps. */

	gnome_init(&argc, &argv);

	/* Store program name for error reports */
	progname = argv[0];

	/* Make the windows */

	make_widgets();

	/* Set initial values for various options */

	now = time(NULL);
	seed = now;
	srand(seed);

	op = '+';
	less_than = 100;
	reset_score(right_pct, NULL);	/* The first arguement is meaningless */

	/* Show it */

	gtk_widget_show(main_window);

	/*
		Run it.
		Exit when the Quit button is pressed.
	*/
	gtk_main();
	return(0);
}

